﻿using Endjin.Assembly.ChangeDetection.Infrastructure;
using System;
using System.Linq;
using System.Reflection;
using System.Runtime;

class Program
{
    [LoaderOptimization(LoaderOptimization.MultiDomain)]
    static public void Main(string[] args)
    {
        for (int i = 0; i < 10000; i++)
        {
            // To load our assembly appdomain neutral we need to use MultiDomain on our hosting and child domain
            // If not we would get different Method tables for the same types which would result in InvalidCastExceptions
            // for the same type.
            var other = AppDomain.CreateDomain("Test" + i.ToString(), AppDomain.CurrentDomain.Evidence, new AppDomainSetup
            {
            //   LoaderOptimization = LoaderOptimization.MultiDomain,
            });

            // Create gate object in other appdomain
            DomainGate gate = (DomainGate)other.CreateInstanceAndUnwrap(Assembly.GetExecutingAssembly().FullName, typeof(DomainGate).FullName);

            // now lets create some data
            CrossDomainData data = new CrossDomainData();
            data.Input = Enumerable.Range(0, 10).ToList();

            // process it in other AppDomain
            DomainGate.Send(gate, data);

            // Display result calculated in other AppDomain
            Console.WriteLine("Calculation in other AppDomain got: {0}", data.Aggregate);
        }
    }
}

class DomainGate : MarshalByRefObject
{
    /// <summary>
    /// Operate on a plain object which is shared from another AppDomain.
    /// </summary>
    /// <param name="gcCount">Total number of GCs</param>
    /// <param name="objAddress">Address to managed object.</param>
    public void DoSomething(int gcCount, IntPtr objAddress)
    {
        if (gcCount != ObjectAddress.GCCount)
        {
            throw new NotSupportedException("During the call a GC did happen. Please try again.");
        }

        // If you get an exception here disable under Projces/Debugging/Enable Visual Studio Hosting Process
        // The appdomain which is used there seems to use LoaderOptimization.SingleDomain
        CrossDomainData data = (CrossDomainData)PtrConverter<Object>.ConvertFromIntPtr(objAddress);

        // process input data from other domain
        foreach (var x in data.Input)
        {
            Console.WriteLine(x);
        }

        // OtherAssembliesUsage user = new OtherAssembliesUsage();

        // generate output data 
        data.Aggregate = data.Input.Aggregate((x, y) => x + y);
    }

    public static void Send(DomainGate gate, object o)
    {
        var old = GCSettings.LatencyMode;
        try
        {
            GCSettings.LatencyMode = GCLatencyMode.Batch; // try to keep the GC out of our stuff
            var addandGCCount = ObjectAddress.GetAddress(o);
            gate.DoSomething(addandGCCount.Value, addandGCCount.Key);
        }
        finally
        {
            GCSettings.LatencyMode = old;
        }
    }
}